//
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// :::::::::::::::::::::::::::::: RadialPong v1.03 ::::::::::::::::::::::::::::::
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//
//
// Disclaimer: I am providing this source code "as-is," without accepting any
//             responsibility for the effects it may have on your computer,
//             network, or livelihood.  Because I am not a full time provider
//             of free Java source code, I cannot offer you any significant
//             technical support.  If you decide to use portions of my code
//             in your project or product, I would be very pleased to be 
//             attributed.  More importantly, please e-mail me to tell me
//             about it!  Do not charge money for this source code, as it is
//             intended to be freely distributed.  Do not remove this header
//             text from the source file, and do not change the version string.
//
// Author: Andrzej Kapolka, ey6es@alumni.cse.ucsc.edu
//
// Description: A paddle game played on a circular track--each paddle is an
//              arc of the circle.  Getting six consecutive wins does
//              "something special."  Player color, computer color, and
//              background color are specifiable as parameters.  Compile under
//              Java version 1.1 or greater.
//
// Easy modifications: There are many constants, defined near the top of the
//                     class definition, that you may modify to affect game
//                     appearance and game play.  To enable the 'secret page'
//                     feature, set secretFeatureEnabled to 'true' and change
//                     the message initialization in init() and the URL
//                     object constructor in mouseDown().  
//


import java.awt.*;
import java.applet.*;
import java.net.*;


public class RadialPong extends Applet implements Runnable
  {
    Thread timer = null;

    // Version
    String myVersion = new String("RadialPong v1.03, by Andrzej Kapolka");

    // Sound objects
    AudioClip hitSound, outSound;

    // Player/computer position in degrees (represents center)
    float playerPos = 180;
    float compPos = 0;

    // The computer's velocity/max velocity (in degrees/frame)
    float compVelocity = 0.0f;
    float compMaxVelocity = 3.0f;

    // Player/computer lengths in degrees
    int playerLength = 23;
    int compLength = 23;

    // Ball width and height
    int ballSize = 6;

    // Ball pos vector and last pos
    int ballX, ballY, ballLastX, ballLastY, ballLastLastX, ballLastLastY;
    float ballDirX, ballDirY;

    // Color constants
    Color playerColor = new Color(0xC8E8FF);
    Color compColor = new Color(0xC0C0FF);
    Color bgColor = new Color(0x000000);

    // Determines speed
    int fps = 60;

    // Determines ball's initial speed
    float ballInitSpeed = 4.0f;

    // Number of movements in rollover
    int rollMoves = 64;

    // Is the 'secret page' feature enabled?
    boolean secretFeatureEnabled = false;
    
    // Roll to/displaying secret screen?
    boolean secretMsg = false;

    // 0=no, 1=player, 2=computer
    int ballAttach = 1;

    // countdown for rollover
    int rollCount = 0;

    int playerWins = 0;
    int subsequentWins = 0;
    int compWins = 0;

    // pos, width and height of track
    int trackX, trackY, trackWidth, trackHeight;

    Color ballColor;
     
    Image offFrame, oldFrame, secretFrame;
    Graphics offG, oldG;

    public void init()
      {
         if(getParameter("playerColor")!=null) {
              playerColor = Color.decode(getParameter("playerColor"));          
         }

         if(getParameter("compColor")!=null) {
              compColor = Color.decode(getParameter("compColor"));          
         }

         if(getParameter("bgColor")!=null) {
              bgColor = Color.decode(getParameter("bgColor"));
         }

         setBackground(bgColor);

         // Load audio files
 
         showStatus("Loading sounds...");

         try {
              hitSound = getAudioClip(new URL(getDocumentBase(),"rpHit.au"));
              outSound = getAudioClip(new URL(getDocumentBase(),"rpOut.au"));
	    } catch(MalformedURLException e) { }

         hitSound.play(); hitSound.stop();
         outSound.play(); outSound.stop();

         offFrame = createImage(size().width,size().height); 
         offG = offFrame.getGraphics();

         offG.setFont(new Font("SansSerif",Font.BOLD,10));

         oldFrame = createImage(size().width, size().height);
         oldG = oldFrame.getGraphics();

         // Set up secret screen
         secretFrame = createImage(size().width, size().height);
         Graphics secG = secretFrame.getGraphics();

         secG.setColor(playerColor);
         secG.fillRect(0,0,size().width,size().height);

         secG.setFont(new Font("SansSerif",Font.BOLD,12));
	    secG.setColor(bgColor);

         int h = secG.getFontMetrics().getHeight();
         int begY = (size().height-(h*3))/2;
         int w = secG.getFontMetrics().stringWidth("*** SECRET ACCESS GRANTED ***");
	    int w2 = secG.getFontMetrics().stringWidth("Click here to access the hidden page:");
	    int w3 = secG.getFontMetrics().stringWidth("'e y 6 e s : (( pastblaster )) !'");
	    secG.drawString("*** SECRET ACCESS GRANTED ***",(size().width-w)/2,begY);
         secG.drawString("Click here to access the hidden page:",(size().width-w2)/2,begY+h*2);
	    secG.drawString("'e y 6 e s : (( pastblaster )) !'",(size().width-w3)/2,begY+h*3);
	   
	    // Center the track       
         trackWidth = size().width-6; trackHeight = size().height-6;
         trackX = trackWidth/2+3; trackY = trackHeight/2+3;
      }

     public void start()
       {
         // Init ball position
         ballAttach = 1;
         ballColor = playerColor;
         double playerTmp = (playerPos/180)*Math.PI;
         ballLastLastY = ballLastY = ballY = (int)((0-((trackWidth-ballSize)/2)*
                                             (double)Math.sin(playerTmp))+trackY);
         ballLastLastX = ballLastX = ballX = (int)(((trackWidth-ballSize)/2)*
                                             (double)Math.cos(playerTmp)+trackX);

         // Start the fps timer
         if(timer==null)
          {
             timer = new Thread(this);
             timer.start();
          }
       }

     public void stop()
       {
         timer = null;
       }

     public void run()
       { 
         showStatus(myVersion);
         
         while(timer!=null)
           {
              try { Thread.sleep(1000/fps); }
              catch(InterruptedException e) { }
              repaint();
           }
       }

    public boolean mouseMove(Event e, int x, int y)
      {
	   // Update player's position
        double playerRad=Math.atan2((double)(x-trackX),(double)(trackY-y));
        playerPos = 90-(float)((playerRad/Math.PI)*180);

        return true;
      }

    public boolean mouseDown(Event e, int x, int y)
      {
        // Release the ball if it's attached to the player
	   if(secretMsg&&rollCount==0) 
	     try {
             getAppletContext().showDocument(new URL(getDocumentBase(),
                                                     "pastblaster/index.html"),
                                            "e y 6 e s : (( pastblaster ))");
		   secretMsg=false;
		   
		   offFrame.getGraphics().drawImage(secretFrame,0,0,this);
		   
		   rollCount = rollMoves;
	     } catch(MalformedURLException yourmom) { }
        else if(!secretMsg&&ballAttach==1)
          {
             ballAttach=0;
             
             double tmpRad = (playerPos/180)*Math.PI;
             ballDirX=(int)((float)(0-Math.cos(tmpRad))*ballInitSpeed);
             ballDirY=(int)((float)Math.sin(tmpRad)*ballInitSpeed);
          }
        return true;
      }
	
    public synchronized void update(Graphics g)
      {
        String pScore = Integer.toString(playerWins), 
               cScore = Integer.toString(compWins);

        // Freeze the last screen, if we're to start rolling
        if(rollCount==rollMoves) oldG.drawImage(offFrame,0,0,this);

        // Clear
        offG.setColor(bgColor); 
        offG.fillRect(0,0,size().width,size().height);

        // Draw scores in upper left and right corners
        offG.setColor(playerColor);
        offG.drawString(pScore,5,offG.getFontMetrics().getHeight());
        offG.setColor(compColor);
        offG.drawString(cScore,size().width-offG.getFontMetrics().stringWidth(cScore)-5,
                        offG.getFontMetrics().getHeight());

        // Draw ball w/trail
        int offset = ballSize/2-2;

        offG.setColor(new Color((int)(ballColor.getRed()*0.25f+bgColor.getRed()*0.75f),
                                (int)(ballColor.getGreen()*0.25f+bgColor.getGreen()*0.75f),
                                (int)(ballColor.getBlue()*0.25f+bgColor.getBlue()*0.75f)));

        offG.fillOval(ballLastLastX-offset-1, ballLastLastY-offset-1, ballSize, ballSize);

        offG.setColor(new Color((int)(ballColor.getRed()*0.5f+bgColor.getRed()*0.5f),
                                (int)(ballColor.getGreen()*0.5f+bgColor.getGreen()*0.5f),
                                (int)(ballColor.getBlue()*0.5f+bgColor.getBlue()*0.5f)));

        offG.fillOval(ballLastX-offset-1, ballLastY-offset-1, ballSize, ballSize);

        offG.setColor(ballColor);

        offG.fillOval(ballX-offset-1, ballY-offset-1, ballSize, ballSize); 

        // Now, draw the paddles

        int compLoc = (int)(compPos - (compLength/2));
        int playerLoc = (int)(playerPos - (playerLength/2));
       
        offG.setColor(compColor);
        offG.drawArc(3, 3, trackWidth, trackHeight, compLoc, compLength);
        offG.drawArc(3, 4, trackWidth, trackHeight, compLoc, compLength);
        offG.drawArc(4, 3, trackWidth, trackHeight, compLoc, compLength);
        offG.drawArc(4, 4, trackWidth, trackHeight, compLoc, compLength);

        offG.setColor(playerColor);
        offG.drawArc(3, 3, trackWidth, trackHeight, playerLoc, playerLength);
        offG.drawArc(3, 4, trackWidth, trackHeight, playerLoc, playerLength);
        offG.drawArc(4, 3, trackWidth, trackHeight, playerLoc, playerLength);
        offG.drawArc(4, 4, trackWidth, trackHeight, playerLoc, playerLength);

        // Paint the image if not rolling
        if(rollCount==0&&!secretMsg) g.drawImage(offFrame,0,0,this);
        else if(rollCount==0&&secretMsg) g.drawImage(secretFrame,0,0,this);
        else
          {
             float picOffset = size().height*(float)rollCount/(float)rollMoves;

             // Draw the old one above
             g.drawImage(oldFrame,0,(int)(picOffset-size().height),this);

             // And the new one below
             if(!secretMsg) g.drawImage(offFrame,0,(int)picOffset,this);
		   else g.drawImage(secretFrame,0,(int)picOffset,this);

             rollCount--;
          }

        // Update ball position

        ballLastLastX = ballLastX;
        ballLastLastY = ballLastY;
        ballLastX = ballX;
        ballLastY = ballY;

        if(ballAttach==0)
          {
             // Move ball by its vector
	        ballX+=(int)ballDirX;
             ballY+=(int)ballDirY;

             // If ball is beyond track, reset
             if(ballX<(0-2*ballSize)||ballY<(0-2*ballSize)||
                ballX>(size().width+2*ballSize)||ballY>(size().height+2*ballSize))
                {
		         outSound.play();

                   // Update scores
			    if(ballColor==playerColor)
                      {
                         playerWins++;
                         subsequentWins++;

                         if(subsequentWins==6 && secretFeatureEnabled)
                              secretMsg = true;

                         ballAttach = 1;

				     // Obviously it's not hard enough, so...
				     if(playerWins<10) 
				        {
				           compMaxVelocity+=0.2f;
				           ballInitSpeed+=0.125f;
				        }
                      }
                   else
                      {
                         compWins++;
                         if(subsequentWins<6) subsequentWins = 0;
                         ballAttach = 2;
                      }

		         // Show version in status window
		         showStatus(myVersion);

                   // Start rolling the screen (if it's not already)
                   if(rollCount==0) rollCount = rollMoves;
                }

              // If ball hit a paddle, change its vector
              double tmpDist = Math.sqrt((double)((trackX-ballX)*(trackX-ballX)+
                               (trackY-ballY)*(trackY-ballY)))+(double)ballSize/2-2;
              double tmpDist2 = Math.sqrt((double)((trackX-ballLastX)*(trackX-ballLastX)+
                               (trackY-ballLastY)*(trackY-ballLastY)))+(double)ballSize/2-3;
              double tmpRad = (double)trackWidth/2;
 
              if(tmpDist>=tmpRad&&tmpDist2<tmpRad)
                {
                   // It's at the right distance, so see if there's a paddle
                   double ballRad = Math.PI/2 - Math.atan2((double)(ballX-trackX),
                                                           (double)(trackY-ballY));
                   double ballPos = (double)((ballRad/Math.PI)*180);

                   double hitAngle = Math.PI/2 - Math.atan2((double)ballDirX,
                                                            (double)(0-ballDirY));

                   double distP = ballPos-playerPos;
                   if(distP>180) distP-=360;
                   else if(distP<-180) distP+=360;

                   double distC = ballPos-compPos;
                   if(distC>180) distC-=360;
                   else if(distC<-180) distC+=360;

                   if(Math.abs(distP)<=(playerLength/2+ballSize/2))
                     {
                        // Player hits ball

				    hitSound.play();

                        ballColor = playerColor;
                             
                        // Find out new direction
                        double playerRad = (playerPos/180)*Math.PI;
                        double tmpAng = (Math.PI+ballRad)+(ballRad-hitAngle)+
                                        (distP/180)*Math.PI;

                        ballDirX=(float)Math.cos(tmpAng)*ballInitSpeed;
                        ballDirY=(float)(0-Math.sin(tmpAng))*ballInitSpeed;
                        ballX+=ballDirX;
                        ballY+=ballDirY;
                     }
                   else if(Math.abs(distC)<=(compLength/2+ballSize/2))
                     {
                        // Computer hits ball

		              hitSound.play();

                        ballColor = compColor;

                        // Find out new direction
				    double compRad = (compPos/180)*Math.PI;
                        double tmpAng = (Math.PI+ballRad)+(ballRad-hitAngle)+
                                        (distC/180)*Math.PI;

                        ballDirX=(float)Math.cos(tmpAng)*ballInitSpeed;
                        ballDirY=(float)(0-Math.sin(tmpAng))*ballInitSpeed;
                        ballX+=ballDirX;
                        ballY+=ballDirY;
                     }

	              // Show version in status window
                   showStatus(myVersion);
                }
          }
        else
          {
             // Put ball on paddle

             if(ballAttach==1)
               {
                   double playerRad = (playerPos/180)*Math.PI;
                   ballY = (int)((0-((trackWidth-ballSize)/2)*
                                     (double)Math.sin(playerRad))+trackY);
                   ballX = (int)(((trackWidth-ballSize)/2)*
                                  (double)Math.cos(playerRad)+trackX);
               }
             else
               {
                   double compRad = (compPos/180)*Math.PI;
                   ballY = (int)((0-((trackWidth-ballSize)/2)*
                                  (double)Math.sin(compRad))+trackY);
                   ballX = (int)(((trackWidth-ballSize)/2)*
                                  (double)Math.cos(compRad)+trackX);
               }
          }
             
        // Update computer's position

        compPos+=compVelocity;
        if(compPos>=360) compPos-=360;
        else if(compPos<0) compPos+=360;

        double a = (double)ballDirX;
        double b = (double)(0-ballDirY);             
        double c = (double)(ballX-trackX);
        double d = (double)(trackY-ballY);
        double e = (double)(trackWidth/2);

        // Distance and direction to travel
        double dist = 0;

        if(ballAttach==0) 
	     {
             double m = -0.5*((Math.sqrt((4*((a*((d*((2*c*b)-(a*d)))+(a*(e*e))))+((b*b)*
                              ((e*e)-(c*c))))))*-1)+(2*c*a)+(2*d*b))/((a*a)+(b*b));

		   if(m>0&&!Double.isNaN(m)&&!Double.isInfinite(m))
		    {
		      double ballTarget = Math.PI/2 - Math.atan2(c+(a*m),d+(b*m));

		      double ballTargetDeg = (ballTarget/Math.PI)*180;

		      if(ballColor==playerColor)
                  {
                     // Shortest distance between paddle and ball's projected target
                     dist = (double)(ballTargetDeg-compPos);              
                  }
		      else if(ballColor==compColor)
                  {
                     // Shortest distance to opposite of ball's projected target
                     dist = (double)((180+ballTargetDeg)-compPos);
                  }
		    }
	     }
	 else
	     {
              double ballRad = Math.PI/2 - Math.atan2(c,d);
                                                     
              double ballPos = (double)((ballRad/Math.PI)*180);

              if(ballAttach==1)
                {
                  // We're not worried yet, so just stay opposite the ball
                  dist = (double)((180+ballPos)-compPos);  
                }
               else if(ballAttach==2)
                {
                  // Release ball after a while (but not while rolling)
                  if((int)(Math.random()*fps*4)==1&&rollCount==0)
                    {
                       ballAttach=0;
             
                       double tmpRad = (compPos/180)*Math.PI;
                       ballDirX=(float)(0-Math.cos(tmpRad))*ballInitSpeed;
                       ballDirY=(float)Math.sin(tmpRad)*ballInitSpeed;
                    }
                  else 
                    {
                       // Change direction once in a while
                       if((int)(Math.random()*fps)==1)
                         {
                             compVelocity = (float)(2*Math.random()*compMaxVelocity)
                                                    -compMaxVelocity;
                         }
                    }
                } 
             }

             if(dist!=0)
               {
                  // Normalize distance
                  if(dist>180) dist-=360;
                  else if(dist<-180) dist+=360;

                  if(Math.abs(dist)<=2) compVelocity = 0;
                  else if(dist>=90) compVelocity=compMaxVelocity;
                  else if(dist<=-90) compVelocity=-1*compMaxVelocity; 
                  else compVelocity = (float)((dist/180)*compMaxVelocity*2);
               }
      }
  }